#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <std.h>
#include <tos.h>
#include <atarierr.h>
#include <scsi3.h>
#include <scsidrv/scsidefs.h>


const char *DEVICE_TYPES[] = {
	"Direct Access",
	"Sequential Access",
	"Printer",
	"Processor",
	"Write-Once",
	"CD-ROM/DVD/BD/DVD-RAM",
	"Scanner",
	"Optical Memory",
	"Medium Changer",
	"Communications",
	"Graphic Arts Pre-Press",
	"Graphic Arts Pre-Press",
	"Storage Array Controller",
	"Enclosure Services",
	"Simplified Direct Access",
	"Optical Card Reader/Writer",
	"Bridge Controller",
	"Object-based Storage",
	"Automation/Drive Interface",
	"Security Manager",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Unknown Device Type",
	"Well Known Logical Unit"
};


typedef struct deviceInfo {
	UWORD busNo;
	UWORD id;
	ULONG maxLen;
	char busName[20];
	char deviceBusName[20];
} DEVICEINFO;


UWORD findDevices(void);
void testCheckDev(UWORD, UWORD);
UWORD testInquiry(void);
void testOpenClose(UWORD, UWORD, ULONG);
bool testError(UWORD, UWORD);
bool testRequestSense(void);
bool testReadCapacity(ULONG *);
bool testRead(ULONG);
bool testReportLuns(void);
bool testSendDiagnostic(void);
void print(const char *, ...);
bool getCookie(LONG, ULONG *);


DEVICEINFO deviceInfos[32];
FILE *out;
bool hasError = false;
tpScsiCall scsiCall;
tSCSICmd cmd;
SENSE_DATA senseData;


#pragma warn -par
void
main(WORD argc, const char *argv[])
{
	DLONG scsiId;
	ULONG maxLen;
	tHandle handle;
	LONG oldstack = 0;
	UWORD devCount;
	int i;

	getCookie('SCSI', (ULONG *)&scsiCall);
	if(!scsiCall) {
		printf("SCSI Driver not found\n");

		Cconin();

		return;
	}

	out = fopen("HDDRTEST.LOG", "w");
	if(!out) {
		printf("Couldn't open 'HDDRTEST.LOG' for writing\n");

		Cconin();

		return;
	}

	print("Found SCSI Driver version $%04X\n\n", scsiCall->Version);

	cmd.Flags = 0;
	cmd.SenseBuffer = (BYTE *)&senseData;
	cmd.Timeout = 2000;

	if(!Super((void *)1L)) oldstack = Super(0L);

	devCount = findDevices();

	for(i = 0; i < devCount; i++) {
		print("\nTesting bus '%s', device ID %d\n",
			deviceInfos[i].deviceBusName, deviceInfos[i].id);

		testCheckDev(deviceInfos[i].busNo, deviceInfos[i].id);
		testOpenClose(deviceInfos[i].busNo, deviceInfos[i].id, deviceInfos[i].maxLen);

		scsiId.hi = 0;
		scsiId.lo = deviceInfos[i].id;

		handle = (tHandle)scsiCall->Open(deviceInfos[i].busNo, &scsiId,
			&maxLen);
		if(handle <= 0) {
			print("    ERROR: No handle\n");

			hasError = true;
		}
		else {
			UWORD deviceType;

			cmd.Handle = handle;

			deviceType = testInquiry();

			hasError |= !testRequestSense();

			if(deviceType != 0x1f) {
				switch(deviceType) {
					case 0x00:
					case 0x05:
					case 0x07:
						{
							ULONG blockSize;

							hasError |= !testReadCapacity(&blockSize);
							if(blockSize) {
								hasError |= !testRead(blockSize);
							}
						}
						break;

					default:
						break;
				}

/* Can only work if no logfile is written to the tested drives */
/*				hasError |= !testError(deviceInfos[i].busNo, deviceInfos[i].id);*/

				hasError |= !testReportLuns();
/*				hasError |= !testSendDiagnostic(); */
			}

			scsiCall->Close(handle);
		}

		if(hasError) {
			print("ERROR\n");

			hasError = false;
		}
		else {
			print("OK\n");
		}
	}

	if(oldstack) Super((void *)oldstack);

	fclose(out);

	Cconin();
}
#pragma warn .par


UWORD
findDevices()
{
	tBusInfo busInfo;
	LONG busResult;
	UWORD devCount = 0;

	print("Buses:\n");

	busResult = scsiCall->InquireSCSI(cInqFirst, &busInfo);
	while(!busResult) {
		tDevInfo devInfo;
		LONG result;

		result = scsiCall->InquireBus(cInqFirst, busInfo.BusNo, &devInfo);
		while(!result) {
			char deviceBusName[20];
			UWORD features;

			scsiCall->CheckDev(busInfo.BusNo, &devInfo.SCSIId,
				deviceBusName, &features);

			deviceInfos[devCount].busNo = busInfo.BusNo;
			deviceInfos[devCount].id = (UWORD)devInfo.SCSIId.lo;
			deviceInfos[devCount].maxLen = busInfo.MaxLen;
			strcpy(deviceInfos[devCount].busName, busInfo.BusName);	
			strcpy(deviceInfos[devCount].deviceBusName, deviceBusName);	
			devCount++;

			result = scsiCall->InquireBus(cInqNext, busInfo.BusNo, &devInfo);
		}

		print("  ID: %d, Name: '%s', Maximum transfer length: %ld ($%lX)\n",
			busInfo.BusNo, busInfo.BusName, busInfo.MaxLen, busInfo.MaxLen);

		busResult = scsiCall->InquireSCSI(cInqNext, &busInfo);
	}

	return devCount;
}


void
testCheckDev(UWORD busNo, UWORD id)
{
	DLONG scsiId;
	LONG result;
	char name[20];
	UWORD features;

	print("  CheckDev()\n");

	print("    Checking with illegal bus ID\n");

	scsiId.hi = 0;
	scsiId.lo = 0;
	result = scsiCall->CheckDev(31, &scsiId, name, &features);
	if(result != EUNDEV) {
		print("    ERROR: Illegal bus ID 31 was accepted\n");

		hasError = true;
	}
				
	print("    Checking with illegal SCSI ID\n");

	scsiId.hi = 1;
	scsiId.lo = 0;
	result = scsiCall->CheckDev(busNo, &scsiId, name, &features);
	if(result != EUNDEV) {
		print("    ERROR: Illegal SCSI ID hi 1 was accepted\n");

		hasError = true;
	}

	scsiId.hi = 0;
	scsiId.lo = 9;
	result = scsiCall->CheckDev(busNo, &scsiId, name, &features);
	if(result != EUNDEV) {
		print("    ERROR: Illegal SCSI ID lo 9 was accepted\n");

		hasError = true;
	}

	print("    Checking with legal bus and SCSI ID\n");

	scsiId.hi = 0;
	scsiId.lo = id;
	result = scsiCall->CheckDev(busNo, &scsiId, name, &features);
	if(result < 0) {
		print("    ERROR: Legal IDs were rejected with error code %ld\n", result);

		hasError = true;
	}
}


UWORD
testInquiry()
{
	SENSE_BLK Inquiry = {
		0x12, 0x00, 0x00, 0x00, 0x00, (UBYTE)sizeof(INQUIRY_DATA), 0x00, 0x00, 0x00
	};

	LONG status;
	UWORD deviceType;
	INQUIRY_DATA inquiryData;
	char name[25];
	char revision[5];

	print("  INQUIRY\n");


	print("    Calling with legal data\n");

	cmd.Cmd = (void *)&Inquiry;
	cmd.CmdLen = 6;
	cmd.Buffer = &inquiryData;
	cmd.TransferLen = Inquiry.length;

	memset(&inquiryData, 0, sizeof(INQUIRY_DATA));

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		return 0;
	}

	deviceType = inquiryData.deviceType;
	if(inquiryData.deviceType == 0x1f) {
		print("    ERROR: Illegal device type\n");

		return 0;
	}

	strncpy(name, inquiryData.vendor, 24);
	strncpy(revision, inquiryData.revision, 4);
	name[24] = 0;
	revision[4] = 0;
	print("      Device type: %s\n",
		DEVICE_TYPES[inquiryData.deviceType & 0x1f]);
	print("      Device name: '%s'\n", name);
	print("      Firmware revision: '%s'\n", revision);

	print("      SCSI/SPC version: ");
	switch(inquiryData.ANSIVersion) {
		case 0:
			print("-");
			break;

		case 1:
			print("SCSI-1 CCS");
			break;

		case 2:
			print("SCSI-2");
			break;

		case 3:
			print("SPC");
			break;

		default:
			{
				char version[6];
				sprintf(version, "SPC-%d", inquiryData.ANSIVersion - 2);
				print(version);
				break;
			}
	}
	print("\n");

	print("      Additional length: $%02X\n", inquiryData.additionalLength);
	if(inquiryData.additionalLength < 0x1f) {
		print("      ERROR: Additional length must be at least $1F\n");

		hasError = true;
	}

	print("    Calling with non-existing LUN 7\n");

	Inquiry.lun = 7;

	memset(&inquiryData, 0, sizeof(INQUIRY_DATA));

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		hasError = true;
	}
	else if(inquiryData.peripheralQualifier != 0x03 ||
		inquiryData.deviceType != 0x1f) {
		print("    ERROR: Call was not properly rejected\n");

		hasError = true;
	}

	Inquiry.lun = 0;

	return deviceType;
}


bool
testRequestSense()
{
	BYTE RequestSense[6] = { 0x03, 0, 0, 0, sizeof(SENSE_DATA), 0 };
	SENSE_DATA localSenseData;

	LONG status;

	print("  REQUEST SENSE\n");


	print("    Calling REQUEST SENSE for existing LUN 0\n");

	cmd.Cmd = (void *)&RequestSense;
	cmd.CmdLen = (UWORD)sizeof(RequestSense);
	cmd.Buffer = &localSenseData;
	cmd.TransferLen = sizeof(SENSE_DATA);

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		hasError = true;
	}
	else {
		print("      Additional sense length: $%02X\n",
			localSenseData.addSenseLength);
		if(localSenseData.addSenseLength < 0x0b) {
			print("      ERROR: Additional sense length must be at least $0B\n");

			hasError = true;
		}
	}


	print("    Calling REQUEST SENSE for non-existing LUN 7\n");

	RequestSense[1] = 7 << 5;

	memset(&localSenseData, 0, sizeof(SENSE_DATA));

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		hasError = true;
	}

	/* Even though GOOD is returned Sense Key and ASC must be set */
	if(localSenseData.senseKey != 0x05 ||
		localSenseData.addSenseCode != 0x25) {
		print("    ERROR: Call was not properly rejected\n");
		print("      Expected: SenseKey $05 ($%02X), ASC $25 ($%02X)\n",
			localSenseData.senseKey, localSenseData.addSenseCode);

		hasError = true;
	}

	RequestSense[1] = 0;


	print("    Calling REQUEST SENSE again for existing LUN 0\n");

	cmd.Cmd = (void *)&RequestSense;
	cmd.CmdLen = (UWORD)sizeof(RequestSense);
	cmd.Buffer = &localSenseData;
	cmd.TransferLen = sizeof(SENSE_DATA);

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		hasError = true;
	}

	return true;
}


void
testOpenClose(UWORD busNo, UWORD id, ULONG maxLen)
{
	DLONG scsiId;
	tHandle handle[64];
	ULONG len;
	LONG result;
	int i;

	print("  Open/Close()\n");

	scsiId.hi = 0;
	scsiId.lo = id;

	for(i = 0; i < 64; i++) {
		handle[i] = (tHandle)scsiCall->Open(busNo, &scsiId, &len);

		if(len != maxLen) {
			print("    ERROR: Transfer length mismatch: $%lX\n", len);

			hasError = true;

			Cconin();

			break;
		}

		result = (LONG)handle[i];

		switch((WORD)result) {
			case ENHNDL:
				break;

			case EUNDEV:
				print("    Handle %d: EUNDEV\n", i);
				break;

			case EACCDN:
				print("    Handle %d: EACCDN\n", i);
				break;

			default:
				break;
		}

		if(result < 0) {
			print("    Available handles: %d\n", i);

			break;
		}
	}

	while(--i >= 0) {
		result = scsiCall->Close(handle[i]);
		if(result != 0) {
			print("    ERROR: Couldn't close handle %d\n", i);

			hasError = true;
		}
	}
}


bool
testError(UWORD busNo, UWORD id)
{
	DLONG scsiId;
	tHandle handle1, handle2;
	LONG status1, status2;
	ULONG maxLen;

	print("  Error()\n");

	scsiId.hi = 0;
	scsiId.lo = id;

	handle1 = (tHandle)scsiCall->Open(busNo, &scsiId, &maxLen);
	if(handle1 <= 0) {
		goto error;
	}

	handle2 = (tHandle)scsiCall->Open(busNo, &scsiId, &maxLen);
	if(handle2 <= 0) {
		goto error;
	}

	status1 = scsiCall->Error(handle1, cErrRead, 1);
	if(status1 != 0) {
		print("    ERROR: cErrMediach status 1 before change: %ld\n", status1);

		goto error;
	}

	status2 = scsiCall->Error(handle2, cErrRead, 1);
	if(status2 != 0) {
		print("    ERROR: cErrMediach status 2 before change: %ld\n", status2);

		goto error;
	}

	status1 = scsiCall->Error(handle1, cErrWrite, 1);
	if(status1 != 0) {
		print("    ERROR: cErrMediach status 1 after change: %ld\n", status1);

		goto error;
	}

	status2 = scsiCall->Error(handle2, cErrRead, 1);
	if(status2 != 1) {
		print("    ERROR: cErrMediach status 2 after change, first call: %ld\n", status2);

		goto error;
	}

	status2 = scsiCall->Error(handle2, cErrRead, 1);
	if(status2 != 0) {
		print("    ERROR: cErrMediach status 2 after change, second call: %ld\n", status2);

		goto error;
	}

	scsiCall->Close(handle1);
	scsiCall->Close(handle2);

	return true;
 
error:
	scsiCall->Close(handle1);
	scsiCall->Close(handle2);

	return false;
}


bool
testReadCapacity(ULONG *blockSize)
{
	BYTE ReadCapacity10[10] = {
		0x25, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};
	BYTE ReadCapacity16[16] = {
		0x9e, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};
	BYTE Read10[10] = {
		0x28, 0, 0, 0, 0, 0, 0, 0, 1, 0
	};

	LONG status;
	ULONG capacity10[] = { 0L, 0L };
	ULONG capacity16[] = { 0L, 0L, 0L };
	WORD buffer[1024];

	*blockSize = 512;

	if(!(*cmd.Handle & cAllCmds)) {
		return true;
	}

	*blockSize = 0;


	print("  READ CAPACITY\n");


	print("    Reading capacity with READ CAPACITY (10)\n");

	cmd.Cmd = (void *)&ReadCapacity10;
	cmd.CmdLen = (UWORD)sizeof(ReadCapacity10);
	cmd.Buffer = capacity10;
	cmd.TransferLen = sizeof(capacity10);

	status = scsiCall->In(&cmd);
	if(status == 2 && senseData.senseKey == 0x02 &&
		senseData.addSenseCode == 0x3a) {
		print("    No medium inserted\n");

		return true;
	}
	else if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		return false;
	}

	if(!capacity10[0]) {
		print("    ERROR: Wrong capacity '0'\n");

		return false;
	}

	print("      Number of blocks: %ld, Block size: %ld\n",
		capacity10[0] + 1, capacity10[1]);


	print("    Reading capacity with READ CAPACITY (16)\n");

	cmd.Cmd = (void *)&ReadCapacity16;
	cmd.CmdLen = (UWORD)sizeof(ReadCapacity16);
	cmd.Buffer = capacity16;
	cmd.TransferLen = sizeof(capacity16);

	status = scsiCall->In(&cmd);
	if(status) {
		print("      READ CAPACITY (16) is not supported\n");
	}
	else {
		if(!capacity16[0] && !capacity16[1]) {
			print("    ERROR: Wrong capacity '0'\n");

			return false;
		}
	}


	*blockSize = capacity10[1];


	print("    Reading last block (%ld)\n", capacity10[0]);

	cmd.Cmd = (void *)&Read10;
	cmd.CmdLen = (UWORD)sizeof(Read10);
	cmd.Buffer = &buffer;
	cmd.TransferLen = capacity10[1];

	Read10[2] = (capacity10[0] >> 24) & 0xff;
	Read10[3] = (capacity10[0] >> 16) & 0xff;
	Read10[4] = (capacity10[0] >> 8) & 0xff;
	Read10[5] = capacity10[0] & 0xff;

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		return false;
	}


	capacity10[0]++;

	print("    Reading last block + 1 (%ld)\n", capacity10[0]);

	cmd.Cmd = (void *)&Read10;
	cmd.CmdLen = (UWORD)sizeof(Read10);
	cmd.Buffer = &buffer;
	cmd.TransferLen = capacity10[1];

	Read10[2] = (capacity10[0] >> 24) & 0xff;
	Read10[3] = (capacity10[0] >> 16) & 0xff;
	Read10[4] = (capacity10[0] >> 8) & 0xff;
	Read10[5] = capacity10[0] & 0xff;

	memset(&senseData, 0, sizeof(SENSE_DATA));

	status = scsiCall->In(&cmd);
	if(status != 2 || senseData.senseKey != 0x05 ||
		senseData.addSenseCode != 0x21) {
		print("    ERROR: Call was not properly rejected\n");
		print("      Expected: SenseKey $05 ($%02X), ASC $21 ($%02X)\n",
			senseData.senseKey, senseData.addSenseCode);

		return false;
	}

	return true;
}


bool
testRead(ULONG blockSize)
{
	BYTE Read6[6] = { 0x08, 0, 0, 0, 1, 0 };
	BYTE Read10[10] = { 0x28, 0, 0, 0, 0, 0, 0, 0, 1, 0 };
	BYTE Read16[16] = { 0x88, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 };

	LONG status;
	WORD buffer1[1024];
	WORD buffer2[1024];
	WORD buffer3[1025];
	BYTE *ptr1, *ptr2, *ptr3;
	UWORD i;

	ptr1 = (BYTE *)buffer1;
	ptr2 = (BYTE *)buffer2;
	ptr3 = (BYTE *)buffer3;

	print("  READ\n");


	print("    Reading block 0 with READ (6)\n");

	cmd.Cmd = (void *)&Read6;
	cmd.CmdLen = (UWORD)sizeof(Read6);
	cmd.Buffer = &buffer1;
	cmd.TransferLen = blockSize;

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		return false;
	}


	if(*cmd.Handle & cAllCmds) {
		print("    Reading block 0 with READ (10)\n");

		cmd.Cmd = (void *)&Read10;
		cmd.CmdLen = (UWORD)sizeof(Read10);
		cmd.Buffer = &buffer2;
		cmd.TransferLen = blockSize;

		status = scsiCall->In(&cmd);
		if(status) {
			print("    ERROR: Call failed: %ld\n", status);

			return false;
		}

		for(i = 0; i < blockSize; i++) {
			if(ptr1[i] != ptr2[i]) {
				break;
			}
		}

		if(i != blockSize) {
			print("    ERROR: Block contents differ at offset %d\n", i);

			hasError = true;
		}


		print("    Reading block 0 with READ (16)\n");

		cmd.Cmd = (void *)&Read16;
		cmd.CmdLen = (UWORD)sizeof(Read16);
		cmd.Buffer = &buffer2;
		cmd.TransferLen = blockSize;

		status = scsiCall->In(&cmd);
		if(status) {
			print("      READ (16) is not supported\n");
		}
		else {
			for(i = 0; i < blockSize; i++) {
				if(ptr1[i] != ptr2[i]) {
					break;
				}
			}

			if(i != blockSize) {
				print("    ERROR: Block contents differ at offset %d\n", i);

				hasError = true;
			}
		}
	}


	print("    Reading block 0 to odd address\n");

	cmd.Cmd = (void *)&Read6;
	cmd.CmdLen = (UWORD)sizeof(Read6);
	cmd.Buffer = &buffer3;
	((BYTE *)cmd.Buffer)++;
	cmd.TransferLen = blockSize;

	status = scsiCall->In(&cmd);
	if(status) {
		print("    ERROR: Call failed: %ld\n", status);

		hasError = true;
	}

	for(i = 0; i < blockSize; i++) {
		if(ptr1[i] != ptr3[i + 1]) {
			break;
		}
	}

	if(i != blockSize) {
		print("    ERROR: Block contents differ at offset %d\n", i);

		hasError = true;
	}


	print("    Reading block 0 from non-existing LUN 7\n");

	Read6[1] = 7 << 5;

	cmd.Cmd = (void *)&Read6;
	cmd.CmdLen = (UWORD)sizeof(Read6);
	cmd.Buffer = &buffer1;
	cmd.TransferLen = blockSize;

	memset(&senseData, 0, sizeof(SENSE_DATA));

	status = scsiCall->In(&cmd);
	if(status != 2 || senseData.senseKey != 0x05 ||
		senseData.addSenseCode != 0x25) {
		print("    ERROR: Call was not properly rejected\n");
		print("      Expected: SenseKey $05 ($%02X), ASC $25 ($%02X)\n",
			senseData.senseKey, senseData.addSenseCode);

		return false;
	}

	Read6[1] = 0;

	return true;
}


bool
testReportLuns()
{
	BYTE ReportLuns[12] = { 0xa0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

	LONG status;
	ULONG buffer[32];

	if(!(*cmd.Handle & cAllCmds)) {
		return true;
	}

	print("  REPORT LUNS\n");


	cmd.Cmd = (void *)&ReportLuns;
	cmd.CmdLen = (UWORD)sizeof(ReportLuns);
	cmd.Buffer = &buffer;
	cmd.TransferLen = sizeof(buffer);

	memset(&senseData, 0, sizeof(SENSE_DATA));

	status = scsiCall->In(&cmd);
	if(!status) {
		ULONG i;

		print("    Number of LUNs: %ld\n", buffer[0] / 8);

		print("      List of LUNs: ");

		for(i = 0; i < buffer[0] / 8; i++) {
			if(i) {
				print(", ");
			}

			print("%ld", buffer[2 + 2 * i + 1]);
		}

		print("\n");
	}
	else if(status == 2 && senseData.senseKey == 0x05 &&
		senseData.addSenseCode == 0x20) {
		print("    REPORT LUNS is not supported\n");
	}
	else if(status == 2 && senseData.senseKey == 0x02 &&
		senseData.addSenseCode == 0x3a) {
		print("    No medium inserted\n");
	}
	else {
		print("    ERROR: Call was not properly rejected\n");
		print("      Expected: Sense Key $05 ($%02X), ASC $20 ($%02X)\n",
			senseData.senseKey, senseData.addSenseCode);

		return false;
	}
	
	return true;
}


bool
testSendDiagnostic()
{
	BYTE SendDiagnostic[6] = { 0x1d, 0x04, 0, 0, 0, 0 };

	LONG status;

	print("  SEND DIAGNOSTIC\n");


	print("    Default self test\n");

	cmd.Cmd = (void *)&SendDiagnostic;
	cmd.CmdLen = (UWORD)sizeof(SendDiagnostic);
	cmd.Buffer = NULL;
	cmd.TransferLen = 0;

	memset(&senseData, 0, sizeof(SENSE_DATA));

	status = scsiCall->In(&cmd);
	if(status == 2 && senseData.senseKey == 0x05 &&
		senseData.addSenseCode == 0x20) {
		print("      SEND DIAGNOSTIC is not supported\n");
	}
	else if(status) {
		print("      ERROR: Call was not properly rejected\n");
		print("        Expected: Sense Key $05 ($%02X), ASC $20 ($%02X)\n",
			senseData.senseKey, senseData.addSenseCode);

		return false;
	}

	return true;
}


void
print(const char *msg, ...)
{
	va_list args;
	char s[161];

	va_start(args, msg);
	vsprintf(s, msg, args);
	va_end(args);
	printf(s);
	fprintf(out, s);
}


LONG
cookieptr()
{
	return *((LONG *)0x5a0);
}


bool
getCookie(LONG cookie, ULONG *p_value)
{
	LONG *cookiejar = (LONG *)Supexec(cookieptr);

	if(!cookiejar) return false;

	do {
		if(cookiejar[0] == cookie) {
			if (p_value) *p_value = (ULONG)cookiejar[1];
			return true;
		}
		else
			cookiejar = &(cookiejar[2]);
	} while(cookiejar[-2]);

	return false;
}
